"""
Tests for
fitbenchmarking.core.fitting_benchmarking.loop_over_benchmark_problems
"""
from __future__ import (absolute_import, division, print_function)
import inspect
import os
import unittest

from fitbenchmarking import mock_problems
from fitbenchmarking.utils import fitbm_result, exceptions
from fitbenchmarking.core.fitting_benchmarking import \
    loop_over_benchmark_problems
from fitbenchmarking.parsing.parser_factory import parse_problem_file
from fitbenchmarking.utils.options import Options
from fitbenchmarking.jacobian.SciPyFD_2point_jacobian import ScipyTwoPoint

# Defines the module which we mock out certain function calls for
FITTING_DIR = "fitbenchmarking.core.fitting_benchmarking"


# Due to structure of tests, some variables may not be previously defined
# in the init function.
# pylint: disable=attribute-defined-outside-init
def make_fitting_problem(file_name='cubic.dat', minimizers=None):
    """
    Helper function that returns a simple fitting problem
    """
    options = Options()
    if minimizers:
        options.minimizers = minimizers

    bench_prob_dir = os.path.dirname(inspect.getfile(mock_problems))
    fname = os.path.join(bench_prob_dir, file_name)

    fitting_problem = parse_problem_file(fname, options)
    fitting_problem.correct_data()
    jac = ScipyTwoPoint(fitting_problem)
    fitting_problem.jac = jac
    return fitting_problem


class LoopOverBenchmarkProblemsTests(unittest.TestCase):
    """
    loop_over_starting_values tests
    """

    def setUp(self):
        """
        Setting up problem for tests
        """
        self.problem = make_fitting_problem()
        self.options = Options()
        self.options.software = ["scipy"]
        self.scipy_len = len(self.options.minimizers["scipy"])
        bench_prob_dir = os.path.dirname(inspect.getfile(mock_problems))
        self.default_parsers_dir = os.path.join(bench_prob_dir,
                                                "default_parsers")
        self.count = 0
        self.result_args = {'options': self.options,
                            'problem': self.problem,
                            'jac': self.problem.jac,
                            'initial_params': self.problem.starting_values[0],
                            'params': [],
                            'chi_sq': 1}
        self.list_results = [fitbm_result.FittingResult(**self.result_args)
                             for i in range(self.scipy_len)]
        self.individual_problem_results = [
            self.list_results, self.list_results]

    def mock_func_call(self, *args, **kwargs):
        """
        Mock function to be used instead of loop_over_starting_values
        """
        individual_problem_results = \
            self.individual_problem_results[self.count]
        problem_fails = self.problem_fails
        unselected_minimzers = {"scipy": []}
        self.count += 1
        return individual_problem_results, problem_fails, unselected_minimzers

    def shared_tests(self, list_len, expected_problem_fails):
        """
        Shared tests for the `loop_over_starting_values` function

        :param list_len: number of expect fitting results
        :type list_len: int
        :param expected_problem_fails: list of problems which fail
        :type expected_problem_fails: list
        """
        results, failed_problems, unselected_minimzers = \
            loop_over_benchmark_problems(self.problem_group,
                                         self.options)
        assert len(results) == list_len
        assert failed_problems == expected_problem_fails
        for keys, values in unselected_minimzers.items():
            assert keys == "scipy"
            assert values == []

    @unittest.mock.patch('{}.loop_over_starting_values'.format(FITTING_DIR))
    def test_run_multiple_benchmark_problems(self, loop_over_starting_values):
        """
        Checks that all benchmark problems run with no failures
        """
        self.problem_fails = []
        loop_over_starting_values.side_effect = self.mock_func_call
        
        self.problem_group = []
        for file_name in ["cubic.dat", "prob_def_1.txt"]:
            self.problem_group.append(
                os.path.join(self.default_parsers_dir, file_name))
        expected_problem_fails = self.problem_fails
        expected_list_length = len(self.list_results) * 2
        self.shared_tests(expected_list_length, expected_problem_fails)

    @unittest.mock.patch('{}.loop_over_starting_values'.format(FITTING_DIR))
    def test_run_multiple_failed_problems(self, loop_over_starting_values):
        """
        Checks that multiple failed problems are reported correctly
        """
        self.problem_fails = ['Random_failed_problem_1',
                              'Random_failed_problem_2']
        loop_over_starting_values.side_effect = self.mock_func_call
        self.problem_group = [os.path.join(self.default_parsers_dir,
                                           "cubic.dat")]

        expected_problem_fails = self.problem_fails
        expected_list_length = len(self.list_results)
        self.shared_tests(expected_list_length, expected_problem_fails)

    @unittest.mock.patch('{}.loop_over_starting_values'.format(FITTING_DIR))
    def test_check_no_results_produced(self, loop_over_starting_values):
        """
        Checks that multiple failed problems are reported correctly
        """
        loop_over_starting_values.side_effect = self.mock_func_call
        problem_group = [os.path.join(self.default_parsers_dir,
                                      "META.txt")]

        with self.assertRaises(exceptions.NoResultsError):
            _, _, _ = \
                loop_over_benchmark_problems(problem_group,
                                             self.options)


if __name__ == "__main__":
    unittest.main()



"""
This file contains a factory implementation for the parsers.
This is used to manage the imports and reduce effort in adding new parsers.
"""

import os
from importlib import import_module
from inspect import getmembers, isabstract, isclass

from fitbenchmarking.parsing.base_parser import Parser
from fitbenchmarking.utils.exceptions import (MissingSoftwareError,
                                              NoParserError)


# By design the factory parser is a class with only one function
# pylint: disable=too-few-public-methods, no-self-use
class ParserFactory:
    """
    A factory for creating parsers.
    This has the capability to select the correct parser, import it, and
    generate an instance of it.
    Parsers generated from this must be a subclass of base_parser.Parser
    """

    @staticmethod
    def create_parser(filename):
        """
        Inspect the input and create a parser that matches the required file.

        :param filename: The path to the file to be parsed
        :type filename: string

        :return: Parser for the problem
        :rtype: fitbenchmarking.parsing.base_parser.Parser subclass
        """

        with open(filename, 'r') as f:
            lines = f.readline()

        # if there's a SIF file ending, use cutest
        extension = os.path.splitext(filename)[1]
        if "SIF" in extension.upper():
            parser_name = 'cutest'
        else:  # Otherwise, take the first section of text
            parser_name = ''
            for line in lines.strip('#').strip():
                if not line.isalpha():
                    break
                parser_name += line

        module_name = '{}_parser'.format(parser_name.lower())

        try:
            module = import_module('.' + module_name, __package__)
        except ImportError as e:
            full_path = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                     module_name + '.py'))
            if os.path.exists(full_path):
                raise MissingSoftwareError('Requirements are missing for the '
                                           '{} parser: {}'.format(
                                               parser_name, str(e)))

            raise NoParserError('Could not find parser for {}. '
                                'Check the input is correct and try '
                                'again.'.format(filename))

        classes = getmembers(module, lambda m: (isclass(m)
                                                and not isabstract(m)
                                                and issubclass(m, Parser)
                                                and m is not Parser))

        return classes[0][1]


def parse_problem_file(prob_file, options):
    """
    Loads the problem file into a fitting problem using the correct parser.

    :param prob_file: path to the problem file
    :type prob_file: string
    :param options: all the information specified by the user
    :type options: fitbenchmarking.utils.options.Options

    :return: problem object with fitting information
    :rtype: fitbenchmarking.parsing.fitting_problem.FittingProblem
    """
    parser = ParserFactory.create_parser(prob_file)
    with parser(prob_file, options) as p:
        problem = p.parse()

    problem.verify()
    return problem



from setuptools import find_packages, setup
from glob import glob

with open('README.md', encoding="utf-8") as f:
    long_description = f.read()

setup(name='FitBenchmarking',
      description='FitBenchmarking: A tool for comparing fitting software',
      author='FitBenchmarking Team',
      author_email="support@fitbenchmarking.com",
      url='http://fitbenchmarking.com',
      long_description=long_description,
      long_description_content_type="text/markdown",
      license='BSD',
      entry_points={
          "console_scripts": [
              'fitbenchmarking = fitbenchmarking.cli.main:main'
          ]
      },
      packages=find_packages(exclude=('*mock*', '*test*')),
      install_requires=['docutils',
                        'numpy',
                        'matplotlib>=2.0',
                        'scipy>=0.18',
                        'lxml',
                        'pandas>=1.0',
                        'jinja2',
                        'configparser'],
      extras_require={'DFO': ['DFO-LS', 'dfogn'],
                      'SAS': ['sasmodels'],
                      'minuit': ['iminuit<2'],
                      'bumps': ['bumps']},
      zip_safe=False,
      use_scm_version={'fallback_version': '0.1.0'},
      setup_requires=['setuptools_scm'],
      include_package_data=True
)



"""
Implements the base class for the fitting software controllers.
"""

from abc import ABCMeta, abstractmethod
import numpy

from fitbenchmarking.utils.exceptions import ControllerAttributeError, \
    UnknownMinimizerError


class Controller:
    """
    Base class for all fitting software controllers.
    These controllers are intended to be the only interface into the fitting
    software, and should do so by implementing the abstract classes defined
    here.
    """

    __metaclass__ = ABCMeta

    VALID_FLAGS = [0, 1, 2, 3]

    def __init__(self, problem):
        """
        Initialise anything that is needed specifically for the
        software, do any work that can be done without knowledge of the
        minimizer to use, or function to fit, and call
        ``super(<software_name>Controller, self).__init__(problem)``
        (the base class's ``__init__`` implementation).
        In this function, you must initialize the a dictionary,
        ``self.algorithm_type``, such that the **keys** are given by:

        - ``all`` - all minimizers
        - ``ls`` - least-squares fitting algorithms
        - ``deriv_free`` - derivative free algorithms (these are algorithms that
          cannot use derivative information. For example, the
          ``Simplex`` method in ``Mantid`` does not require Jacobians, and so is
          derivative free.
          However, ``lm-scipy-no-jac`` in ``scipy_ls`` is designed to use
          derivatives, but calculates an approximation internally if one is not
          supplied.)
        - ``general`` - minimizers which solve a generic `min f(x)`.

        The **values** of the dictionary are given as a list of minimizers
        for that specific controller that fit into each of the above categories.
        See for example the ``GSL`` controller.

        :param problem: The parsed problem
        :type problem:
                :class:`~fitbenchmarking.parsing.fitting_problem.FittingProblem`
        """

        # Problem: The problem object from parsing
        self.problem = problem

        # Data: Data used in fitting. Might be different from problem
        #       if corrections are needed (e.g. startX)
        self.data_x = problem.data_x
        self.data_y = problem.data_y
        self.data_e = problem.data_e

        # Initial Params: The starting values for params when fitting
        self.initial_params = None
        # Staring Valuess: The list of starting parameters
        self.starting_values = problem.starting_values
        # Parameter set: The index of the starting parameters to use
        self.parameter_set = None
        # Minimizer: The current minimizer to use
        self.minimizer = None

        # Final Params: The final values for the params from the minimizer
        self.final_params = None

        # Flag: error handling flag
        self._flag = None

        # Algorithm check: this is used to check whether the selected
        # minimizer/minimizers from the options is within the softwares
        # algorithms. It also used to filter out algorithms based on the keys
        # of the dictionary
        self.algorithm_check = {'all': [None],
                                'ls': [None],
                                'deriv_free': [None],
                                'general': [None]}

    @property
    def flag(self):
        """
        | 0: `Successfully converged`
        | 1: `Software reported maximum number of iterations exceeded`
        | 2: `Software run but didn't converge to solution`
        | 3: `Software raised an exception`
        """

        return self._flag

    @flag.setter
    def flag(self, value):

        if value not in self.VALID_FLAGS:
            raise ControllerAttributeError(
                'controller.flag must be one of {}. Got: {}.'.format(
                    list(self.VALID_FLAGS), value))
        self._flag = int(value)

    def prepare(self):
        """
        Check that function and minimizer have been set.
        If both have been set, run self.setup().
        """

        if (self.minimizer is not None) and (self.parameter_set is not None):
            self.initial_params = \
                list(self.starting_values[self.parameter_set].values())
            self.setup()
        else:
            raise ControllerAttributeError('Either minimizer or parameter_set '
                                           'is set to None.')

    def eval_chisq(self, params, x=None, y=None, e=None):
        """
        Computes the chisq value

        :param params: The parameters to calculate residuals for
        :type params: list
        :param x: x data points, defaults to self.data_x
        :type x: numpy array, optional
        :param y: y data points, defaults to self.data_y
        :type y: numpy array, optional
        :param e: error at each data point, defaults to self.data_e
        :type e: numpy array, optional

        :return: The sum of squares of residuals for the datapoints at the
                 given parameters
        :rtype: numpy array
        """
        out = self.problem.eval_r_norm(params=params, x=x, y=y, e=e)
        return out

    def validate_minimizer(self, minimizer, algorithm_type):
        """
        Helper function which checks that the selected minimizer from the
        options (options.minimizer) exists and whether the minimizer is in
        self.algorithm_check[options.algorithm_type] (this is a list set in
        the controller)

        :param minimizer: string of minimizers selected from the
                          options
        :type minimizer: str
        :param algorithm_type: the algorithm type selected from the options
        :type algorithm_type: str
        """
        minimzer_selection = self.algorithm_check[algorithm_type]
        result = minimizer in minimzer_selection

        if not result:
            message = 'The minimizer selected, {0}, is not within ' \
                'algorithm_check[options.algorithm_type] = {1}\n'.format(
                    minimizer, minimzer_selection)
            raise UnknownMinimizerError(message)

    def check_attributes(self):
        """
        A helper function which checks all required attributes are set
        in software controllers
        """
        values = {'_flag': int, 'final_params': numpy.ndarray}

        for attr_name, attr_type in values.items():
            attr = getattr(self, attr_name)
            if attr_type != numpy.ndarray:
                if not isinstance(attr, attr_type):
                    raise ControllerAttributeError(
                        'Attribute "{}" in the controller is not the expected '
                        'type. Expected "{}", got {}.'.format(
                            attr_name, attr_type, type(attr)))
            else:
                # Mantid multifit produces final params as a list of final
                # params.
                if not self.problem.multifit:
                    attr = [attr]
                for a in attr:
                    if any(numpy.isnan(n) or numpy.isinf(n) for n in a):
                        raise ControllerAttributeError(
                            'Attribute "{}" in the controller is not the '
                            'expected numpy ndarray of floats. Expected a '
                            'list or numpy ndarray of floats, got '
                            '{}'.format(attr_name, attr))

    def setup(self):
        """
        Setup the specifics of the fitting.

        Anything needed for "fit" that can only be done after knowing the
        minimizer to use and the function to fit should be done here.
        Any variables needed should be saved to self (as class attributes).
        """
        raise NotImplementedError

    @abstractmethod
    def fit(self):
        """
        Run the fitting.

        This will be timed so should include only what is needed to fit the data.
        """
        raise NotImplementedError

    @abstractmethod
    def cleanup(self):
        """
        Retrieve the result as a numpy array and store results.

        Convert the fitted parameters into a numpy array, saved to
        ``self.final_params``, and store the error flag as ``self.flag``.

        The flag corresponds to the following messages:

        .. automethod:: fitbenchmarking.controllers.base_controller.Controller.flag()
                      :noindex:
        """
        raise NotImplementedError



"""
Implements a controller for the Bumps fitting software.
"""

from bumps.fitters import fit as bumpsFit
from bumps.names import Curve, FitProblem

import numpy as np

from fitbenchmarking.controllers.base_controller import Controller


class BumpsController(Controller):
    """
    Controller for the Bumps fitting software.

    Sasview requires a model to fit.
    Setup creates a model with the correct function.
    """

    def __init__(self, problem):
        """
        Extract param names for function setup

        :param problem: Problem to fit
        :type problem: FittingProblem
        """
        super(BumpsController, self).__init__(problem)

        self._param_names = self.problem.param_names

        self._func_wrapper = None
        self._fit_problem = None
        self._bumps_result = None
        self.algorithm_check = {
            'all': ['amoeba', 'lm-bumps', 'newton', 'de', 'mp'],
            'ls': ['lm-bumps', 'mp'],
            'deriv_free': ['amoeba', 'de'],
            'general': ['amoeba', 'newton', 'de']}

    def setup(self):
        """
        Setup problem ready to run with Bumps.

        Creates a FitProblem for calling in the fit() function of Bumps
        """
        # Bumps fails with the *args notation
        param_name_str = ', '.join(self._param_names)
        wrapper = "def fitFunction(x, {}):\n".format(param_name_str)
        wrapper += "    return func(x, {})".format(param_name_str)

        exec_dict = {'func': self.problem.function}
        exec(wrapper, exec_dict)

        model = exec_dict['fitFunction']

        # Remove any function attribute. BinWidth is the only attribute in all
        # FitBenchmark (Mantid) problems.
        param_dict = {name: value
                      for name, value
                      in zip(self._param_names, self.initial_params)}

        # Create a Function Wrapper for the problem function. The type of the
        # Function Wrapper is acceptable by Bumps.
        func_wrapper = Curve(fn=model,
                             x=self.data_x,
                             y=self.data_y,
                             dy=self.data_e,
                             **param_dict)

        # Set a range for each parameter
        val_ranges = self.problem.value_ranges
        for name in self._param_names:
            min_val = -np.inf
            max_val = np.inf
            if val_ranges is not None and name in val_ranges:
                min_val = val_ranges[name][0]
                max_val = val_ranges[name][1]
            func_wrapper.__dict__[name].range(min_val, max_val)

        # Create a Problem Wrapper. The type of the Problem Wrapper is
        # acceptable by Bumps fitting.
        self._func_wrapper = func_wrapper
        self._fit_problem = FitProblem(func_wrapper)
        if self.minimizer == "lm-bumps":
            self.minimizer = "lm"

    def fit(self):
        """
        Run problem with Bumps.
        """
        result = bumpsFit(self._fit_problem, method=self.minimizer)

        self._bumps_result = result
        self._status = self._bumps_result.status

    def cleanup(self):
        """
        Convert the result to a numpy array and populate the variables results
        will be read from.
        """
        if self._status == 0:
            self.flag = 0
        elif self._status == 2:
            self.flag = 1
        else:
            self.flag = 2

        self.final_params = self._bumps_result.x



"""
Test file to test the fitting_problem file.
"""

from collections import OrderedDict
from unittest import TestCase

import numpy as np

from fitbenchmarking.parsing.fitting_problem import FittingProblem
from fitbenchmarking.utils import exceptions
from fitbenchmarking.utils.options import Options


class TestFittingProblem(TestCase):
    """
    Class to test the FittingProblem class
    """

    def setUp(self):
        """
        Setting up FittingProblem tests
        """
        self.options = Options()

    def test_verify_problem(self):
        """
        Test that verify only passes if all required values are set.
        """
        fitting_problem = FittingProblem(self.options)
        with self.assertRaises(exceptions.FittingProblemError):
            fitting_problem.verify()
            self.fail('verify() passes when no values are set.')

        fitting_problem.starting_values = [
            OrderedDict([('p1', 1), ('p2', 2)])]
        with self.assertRaises(exceptions.FittingProblemError):
            fitting_problem.verify()
            self.fail('verify() passes starting values are set.')

        fitting_problem.data_x = np.array([1, 2, 3, 4, 5])
        with self.assertRaises(exceptions.FittingProblemError):
            fitting_problem.verify()
            self.fail('verify() passes when data_x is set.')

        fitting_problem.data_y = np.array([1, 2, 3, 4, 5])
        with self.assertRaises(exceptions.FittingProblemError):
            fitting_problem.verify()
            self.fail('verify() passes when data_y is set.')

        fitting_problem.function = lambda x, p1, p2: p1 + p2
        try:
            fitting_problem.verify()
        except exceptions.FittingProblemError:
            self.fail('verify() fails when all required values set.')

        fitting_problem.data_x = [1, 2, 3]
        with self.assertRaises(exceptions.FittingProblemError):
            fitting_problem.verify()
            self.fail('verify() passes for x values not numpy.')

    def test_eval_f(self):
        """
        Test that eval_f is running the correct function
        """
        fitting_problem = FittingProblem(self.options)
        self.assertRaises(exceptions.FittingProblemError,
                          fitting_problem.eval_f,
                          x=2,
                          params=[1, 2, 3])
        fitting_problem.function = lambda x, p1: x + p1
        x_val = np.array([1, 8, 11])
        eval_result = fitting_problem.eval_f(x=x_val,
                                             params=[5])
        self.assertTrue(all(eval_result == np.array([6, 13, 16])))

        fitting_problem.data_x = np.array([20, 21, 22])
        eval_result = fitting_problem.eval_f(params=[5])
        self.assertTrue(all(eval_result == np.array([25, 26, 27])))

    def test_eval_r(self):
        """
        Test that eval_r is correct
        """
        fitting_problem = FittingProblem(self.options)
        self.assertRaises(exceptions.FittingProblemError,
                          fitting_problem.eval_r,
                          params=[1, 2, 3],
                          x=2)

        fitting_problem.function = lambda x, p1: x + p1
        x_val = np.array([1, 8, 11])
        y_val = np.array([6, 10, 20])

        eval_result = fitting_problem.eval_r(x=x_val,
                                             y=y_val,
                                             params=[5])
        self.assertTrue(all(eval_result == np.array([0, -3, 4])))

        e_val = np.array([2, 4, 1])
        eval_result = fitting_problem.eval_r(x=x_val,
                                             y=y_val,
                                             e=e_val,
                                             params=[5])
        self.assertTrue(all(eval_result == np.array([0, -0.75, 4])))

        fitting_problem.data_x = np.array([20, 21, 22])
        fitting_problem.data_y = np.array([20, 30, 35])
        eval_result = fitting_problem.eval_r(params=[5])
        self.assertTrue(all(eval_result == np.array([-5, 4, 8])))

        fitting_problem.data_e = np.array([2, 5, 10])
        eval_result = fitting_problem.eval_r(params=[5])
        self.assertTrue(all(eval_result == np.array([-2.5, 0.8, 0.8])))

    def test_eval_r_norm(self):
        """
        Test that eval_r_norm is correct
        """
        fitting_problem = FittingProblem(self.options)
        fitting_problem.function = lambda x, p1: x + p1
        x_val = np.array([1, 8, 11])
        y_val = np.array([6, 10, 20])
        e_val = np.array([0.5, 10, 0.1])

        eval_result = fitting_problem.eval_r_norm(params=[5],
                                                  x=x_val,
                                                  y=y_val,
                                                  e=e_val)
        self.assertEqual(eval_result, 1600.09)

        fitting_problem.data_x = x_val
        fitting_problem.data_y = y_val
        eval_result = fitting_problem.eval_r_norm(params=[5])
        self.assertEqual(eval_result, 25)

    def test_eval_starting_params(self):
        """
        Test that eval_starting_params returns the correct result
        """
        fitting_problem = FittingProblem(self.options)
        self.assertRaises(exceptions.FittingProblemError,
                          fitting_problem.eval_starting_params,
                          param_set=0)
        fitting_problem.function = lambda x, p1: x + p1
        fitting_problem.starting_values = [OrderedDict([('p1', 3)]),
                                           OrderedDict([('p1', 7)])]
        fitting_problem.data_x = np.array([1])
        eval_result = fitting_problem.eval_starting_params(0)
        self.assertTrue(all(eval_result == np.array([4])))
        eval_result = fitting_problem.eval_starting_params(1)
        self.assertTrue(all(eval_result == np.array([8])))

    def test_get_function_params(self):
        """
        Tests that the function params is formatted correctly
        """
        fitting_problem = FittingProblem(self.options)
        expected_function_def = 'a=1, b=2.0, c=3.3, d=4.99999'
        fitting_problem.starting_values = [
            OrderedDict([('a', 0), ('b', 0), ('c', 0), ('d', 0)])]
        params = [1, 2.0, 3.3, 4.99999]
        function_def = fitting_problem.get_function_params(params=params)
        self.assertEqual(function_def, expected_function_def)

    def test_correct_data_single_fit(self):
        """
        Tests that correct data gives the expected result
        """
        fitting_problem = FittingProblem(self.options)
        x_data = np.array([-0.5, 0.0, 1.0, 0.5, 1.5, 2.0, 2.5, 3.0, 4.0])
        y_data = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])
        e_data = np.array([1.0, 20.0, 30.0, 40.0, 50.0, 60.0, 70.0, 80.0, 9.0])
        start_x = 0.5
        end_x = 2.5
        expected_x_data = np.array([0.5, 1.0, 1.5, 2.0, 2.5])
        expected_y_data = np.array([3.0, 2.0, 4.0, 5.0, 6.0])
        expected_e_data = np.array([40.0, 30.0, 50.0, 60.0, 70.0])

        fitting_problem.data_x = x_data
        fitting_problem.data_y = y_data
        fitting_problem.data_e = e_data
        fitting_problem.start_x = start_x
        fitting_problem.end_x = end_x

        fitting_problem.correct_data()
        self.assertTrue(
            np.isclose(fitting_problem.data_x[fitting_problem.sorted_index],
                       expected_x_data).all())
        self.assertTrue(
            np.isclose(fitting_problem.data_y[fitting_problem.sorted_index],
                       expected_y_data).all())
        self.assertTrue(
            np.isclose(fitting_problem.data_e[fitting_problem.sorted_index],
                       expected_e_data).all())
        self.options.use_errors = False
        fitting_problem.correct_data()
        self.assertTrue(
            np.isclose(fitting_problem.data_x[fitting_problem.sorted_index],
                       expected_x_data).all())
        self.assertTrue(
            np.isclose(fitting_problem.data_y[fitting_problem.sorted_index],
                       expected_y_data).all())
        self.assertIs(fitting_problem.data_e, None)

    def test_correct_data_multi_fit(self):
        """
        Tests correct data on a multifit problem.
        """
        fitting_problem = FittingProblem(self.options)
        fitting_problem.multifit = True
        x_data = [np.array([-0.5, 0.0, 1.0, 0.5, 1.5, 2.0, 2.5, 3.0, 4.0]),
                  np.array([-0.5, 0.0, 1.0, 0.5, 1.4, 2.0, 2.5, 3.0, 4.0]),
                  np.array([-0.5, 0.0, 1.0, 0.5, 1.7, 2.0, 2.5, 3.0, 4.0])]
        y_data = [np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0]),
                  np.array([0.0, 1.0, 2.0, 3.0, 24.0, 5.0, 6.0, 7.0, 8.0]),
                  np.array([0.0, 1.0, 2.8, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])]
        e_data = [np.array([1.0, 20.0, 30.0, 40.0, 50.0, 60.0, 1.0, 6.0, 9.0]),
                  np.array([1.0, 20.0, 30.0, 40.0, 50.0, 60.0, 1.0, 6.0, 9.0]),
                  np.array([1.0, 20.0, 30.0, 40.0, 50.0, 60.0, 1.0, 6.0, 9.0])]
        start_x = [0.5, 1.1, 0.0]
        end_x = [2.5, 2.6, 1.0]
        expected_x_data = [np.array([0.5, 1.0, 1.5, 2.0, 2.5]),
                           np.array([1.4, 2.0, 2.5]),
                           np.array([0.0, 0.5, 1.0])]
        expected_y_data = [np.array([3.0, 2.0, 4.0, 5.0, 6.0]),
                           np.array([24.0, 5.0, 6.0]),
                           np.array([1.0, 3.0, 2.8])]
        expected_e_data = [np.array([40.0, 30.0, 50.0, 60.0, 1.0]),
                           np.array([50.0, 60.0, 1.0]),
                           np.array([20.0, 40.0, 30.0])]

        fitting_problem.data_x = x_data
        fitting_problem.data_y = y_data
        fitting_problem.data_e = e_data
        fitting_problem.start_x = start_x
        fitting_problem.end_x = end_x

        fitting_problem.correct_data()
        for i in range(3):
            self.assertTrue(
                np.isclose(
                    fitting_problem.data_x[i][fitting_problem.sorted_index[i]],
                    expected_x_data[i]).all())
            self.assertTrue(
                np.isclose(
                    fitting_problem.data_y[i][fitting_problem.sorted_index[i]],
                    expected_y_data[i]).all())
            self.assertTrue(
                np.isclose(
                    fitting_problem.data_e[i][fitting_problem.sorted_index[i]],
                    expected_e_data[i]).all())

        self.options.use_errors = False
        fitting_problem.correct_data()
        for i in range(3):
            self.assertTrue(
                np.isclose(
                    fitting_problem.data_x[i][fitting_problem.sorted_index[i]],
                    expected_x_data[i]).all())
            self.assertTrue(
                np.isclose(
                    fitting_problem.data_y[i][fitting_problem.sorted_index[i]],
                    expected_y_data[i]).all())
            self.assertIs(fitting_problem.data_e[i], None)



"""
Miscellaneous functions and utilities used in fitting benchmarking.
"""

from __future__ import absolute_import, division, print_function

import glob
import os

from fitbenchmarking.utils.exceptions import NoDataError
from fitbenchmarking.utils.log import get_logger

LOGGER = get_logger()

def get_problem_files(data_dir):
    """
    Gets all the problem definition files from the specified problem
    set directory.

    :param data_dir: directory containing the problems
    :type data_dir: str 

    :return: array containing of paths to the problems
             e.g. In NIST we would have
             [low_difficulty/file1.txt, ..., ...]
    :rtype: list of str
    """

    test_data = glob.glob(data_dir + '/*.*')
    if test_data == []:
        raise NoDataError('"{}" not recognised as a dataset. '
                          'Check that it contains problem files '
                          'and try again.'.format(data_dir))
    problems = [os.path.join(data_dir, data)
                for data in test_data
                if not data.endswith('META.txt')]
    problems.sort()
    for problem in problems:
        LOGGER.debug(problem)

    return problems

def get_css(options,working_directory):
    """
    Returns the path of the local css folder
    
    :param working_directory: location of current directory
    :type working_directory: string
    
    :return: A dictionary containing relative links to the local css directory
    :rtype: dict of strings
    """
    local_css_dir = os.path.join(options.results_dir,"css")
    css_path = os.path.relpath(local_css_dir,working_directory)
    css_dict = {
        'main'   : os.path.join(css_path,'main_style.css'),
        'table'  : os.path.join(css_path,'table_style.css'),
        'custom' : os.path.join(css_path,'custom_style.css')
    }
    
    return css_dict

def get_js(options,working_directory):
    """
    Returns the path of the local js folder
    
    :param working_directory: location of current directory
    :type working_directory: string
    
    :return: A dictionary containing relative links to the local js directory
    :rtype: dict of strings
    """
    local_js_dir = os.path.join(options.results_dir,"js")
    js_path = os.path.relpath(local_js_dir,working_directory)
    js_dict = {
        'mathjax'   : os.path.join(js_path,'tex-mml-chtml.js'),
    }
    
    return js_dict



"""
Implements the base class for the Jacobian.
"""
from abc import ABCMeta, abstractmethod

import numpy as np
from scipy.optimize._numdiff import approx_derivative


class Jacobian:
    """
    Base class for Jacobian.
    """
    __metaclass__ = ABCMeta

    def __init__(self, problem):
        self.problem = problem

    @abstractmethod
    def eval(self, params, func=None, **kwargs):
        """
        Evaluates Jacobian

        :param params: The parameter values to find the Jacobian at
        :type params: list
        :param func: Function to find the Jacobian for, defaults to
                     problem.eval_r
        :type func: Callable, optional

        :return: Approximation of the Jacobian
        :rtype: numpy array
        """
        return NotImplementedError



"""
Utility functions to support logging for the fitbenchmarking
project.
"""
import logging
import sys


def setup_logger(log_file='./fitbenchmarking.log', name='fitbenchmarking',
                 append=False, level='INFO'):
    """
    Define the location and style of the log file.

    :param log_file: path to the log file, defaults to './fitbenchmarking.log'
    :type log_file: str, optional
    :param name: The name of the logger to run the setup for,
                 defaults to fitbenchmarking
    :type name: str, optional
    :param append: Whether to append to the log or create a new one,
                   defaults to False
    :type append: bool, optional
    :param level: The level of error to print, defaults to 'INFO'
    :type level: str, optional
    """
    FORMAT = '[%(asctime)s]  %(levelname)s %(filename)s: %(message)s'
    formatter = logging.Formatter(FORMAT, "%H:%M:%S")

    handler = logging.FileHandler(log_file, mode='a' if append else 'w')
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    for h in logger.handlers:
        logger.removeHandler(h)

    logger.addHandler(handler)

    # Define a Handler which writes <level> or higher messages to console
    levels = {'CRITICAL': logging.CRITICAL,
              'ERROR': logging.ERROR,
              'WARNING': logging.WARNING,
              'INFO': logging.INFO,
              'DEBUG': logging.DEBUG,
              'NOTSET': logging.NOTSET}
    log_level = levels.get(level.upper(), logging.INFO)
    console = logging.StreamHandler(sys.stdout)
    console.setLevel(log_level)
    logger.addHandler(console)
    logger.propagate = False


def get_logger(name='fitbenchmarking'):
    """
    Get the unique logger for the given name.
    This is a straight pass through but will be more intutive for people who
    have not used python logging.

    :param name: Name of the logger to use, defaults to 'fitbenchmarking'
    :type name: str, optional

    :return: The named logger
    :rtype: logging.Logger
    """
    return logging.getLogger(name)



from __future__ import (absolute_import, division, print_function)

import numpy as np
import os
try:
    from tempfile import TemporaryDirectory
except ImportError:
    from backports.tempfile import TemporaryDirectory
import unittest

from fitbenchmarking.parsing.fitting_problem import FittingProblem
from fitbenchmarking.jacobian.SciPyFD_2point_jacobian import ScipyTwoPoint
from fitbenchmarking.utils.fitbm_result import FittingResult
from fitbenchmarking.utils.options import Options
from fitbenchmarking.results_processing import plots


class PlotTests(unittest.TestCase):
    """
    Test the plot object is correct.
    """

    def setUp(self):
        self.opts = Options()
        self.opts.use_errors = True

        self.prob = FittingProblem(self.opts)
        self.prob.data_x = np.array([1, 2, 4, 3, 5])
        self.prob.sorted_index = np.array([0, 1, 3, 2, 4])
        self.prob.data_y = np.array([4, 3, 5, 2, 1])
        self.prob.data_e = np.array([0.5, 0.2, 0.3, 0.1, 0.4])
        self.prob.starting_values = [{'x': 1, 'y': 2}]
        self.prob.eval_f = lambda x, y: x[0] * y + x[1]
        self.prob.name = 'full name'
        jac = ScipyTwoPoint(self.prob)
        self.fr = FittingResult(options=self.opts,
                                problem=self.prob,
                                jac=jac,
                                chi_sq=1.0,
                                initial_params=[1.8],
                                params=[1.2],
                                runtime=2.0,
                                minimizer='fit',
                                error_flag=1)

        self.opts = Options()
        self.opts.use_errors = True

        self.dir = TemporaryDirectory()
        self.plot = plots.Plot(best_result=self.fr,
                               options=self.opts,
                               figures_dir=self.dir.name)

    def test_init_creates_line(self):
        """
        Tests that init puts a single line in the plot.
        """
        lines = self.plot.ax.get_lines()
        self.assertEqual(len(lines), 1)

    def test_plot_data_creates_line(self):
        """
        Test that plotting a line when line_plot is None creates a new line.
        """
        self.plot.line_plot = None
        lines = self.plot.ax.get_lines()
        lines_before = len(lines)
        self.plot.plot_data(False, self.plot.ini_guess_plot_options)
        lines = self.plot.ax.get_lines()
        self.assertEqual(lines_before + 1, len(lines))

    def test_plot_data_updates_line(self):
        """
        Tests that plotting when line_plot is not None updates the line.
        """
        self.plot.plot_data(False,
                            self.plot.ini_guess_plot_options,
                            [1, 2, 3],
                            [2, 3, 1])
        lines = self.plot.ax.get_lines()
        lines_before = len(lines)
        self.plot.plot_data(False, self.plot.ini_guess_plot_options)
        lines = self.plot.ax.get_lines()
        self.assertEqual(lines_before, len(lines))

    def test_plot_initial_guess_create_files(self):
        """
        Test that initial plot creates a file.
        """
        file_name = self.plot.plot_initial_guess()

        self.assertEqual(file_name, 'start_for_full_name.png')
        path = os.path.join(self.dir.name, file_name)
        self.assertTrue(os.path.exists(path))

    def test_plot_best_create_files(self):
        """
        Test that best plot creates a file.
        """
        file_name = self.plot.plot_best('fit', [0.1, 3])

        self.assertEqual(file_name, 'fit_fit_for_full_name.png')
        path = os.path.join(self.dir.name, file_name)
        self.assertTrue(os.path.exists(path))

    def test_plot_fit_create_files(self):
        """
        Test that fit plot creates a file.
        """
        file_name = self.plot.plot_fit('fit', [8, 6.2])

        self.assertEqual(file_name, 'fit_fit_for_full_name.png')
        path = os.path.join(self.dir.name, file_name)
        self.assertTrue(os.path.exists(path))


if __name__ == "__main__":
    unittest.main()



'''
Test the options write function
'''
import copy
import os
import unittest

from fitbenchmarking.utils import exceptions
from fitbenchmarking.utils.options import Options


class OptionsWriteTests(unittest.TestCase):
    """
    Tests for options.write
    """

    def setUp(self):
        '''
        Create an options file and store input
        '''
        config_str = """
            [MINIMIZERS]
            scipy: CG
                   Powell
            dfo: dfogn

            [FITTING]
            algorithm_type: all
            use_errors: no
            num_runs: 2
            software: scipy
                      dfo
            jac_method: SciPyFD
            num_method: cs

            [PLOTTING]
            make_plots: no
            colour_scale: 17.1, b_string?
                          inf, final_string
            comparison_mode: abs
            table_type: acc
                        runtime
            results_dir: new_results

            [LOGGING]
            file_name: THE_LOG.log
            append: yes
            level: DEBUG
            external_output: display
            """

        opts_file = 'test_options_tests.ini'
        with open(opts_file, 'w') as f:
            f.write(config_str)
        self.options_file = opts_file

    def tearDown(self):
        """
        Removes options file after tests are run
        """
        os.remove(self.options_file)

    def test_write(self):
        """
        Test that the options writer works.
        """
        options = Options(file_name=self.options_file)
        new_file_name = 'copy_of_{}'.format(self.options_file)

        options.write(new_file_name)
        new_options = Options(new_file_name)

        os.remove(new_file_name)

        assert options.stored_file_name == self.options_file
        assert new_options.stored_file_name == new_file_name

        # Overwrite file names
        options.stored_file_name = ""
        new_options.stored_file_name = ""

        self.assertDictEqual(options.__dict__, new_options.__dict__)

    def test_write_to_stream(self):
        """
        Test that the stream options writer works.
        """
        options = Options(file_name=self.options_file)
        new_file_name = 'copy_of_{}'.format(self.options_file)

        # open stream, write to it and close it
        f = open(new_file_name, 'w')
        options.write_to_stream(f)
        f.close()

        new_options = Options(new_file_name)

        assert options.stored_file_name == self.options_file
        assert new_options.stored_file_name == new_file_name

        # Overwrite file names
        options.stored_file_name = ""
        new_options.stored_file_name = ""

        os.remove(new_file_name)
        self.assertDictEqual(options.__dict__, new_options.__dict__)

    def test_create_config(self):
        """
        Test that created config object contains all valid sections.
        """
        options = Options(file_name=self.options_file)

        config = options._create_config()
        for section in options.VALID_SECTIONS:
            self.assertTrue(config.has_section(section))

    def test_user_section_valid(self):
        """
        Tests that the user defined sections are correct.
        """
        config_str = """
            [MINIMIZERS]

            [FITTING]
            """
        opts_file = 'test_options_tests_valid.ini'
        with open(opts_file, 'w') as f:
            f.write(config_str)
        Options(opts_file)
        os.remove(opts_file)

    def test_user_section_invalid(self):
        """
        Tests that the user defined sections are incorrect.
        """
        config_str = """
            [MINIMIZERS SECTION]

            [FITTING]
            """
        opts_file = 'test_options_tests_valid.ini'
        with open(opts_file, 'w') as f:
            f.write(config_str)
        with self.assertRaises(exceptions.OptionsError):
            Options(opts_file)
        os.remove(opts_file)

    def test_options_reset(self):
        """
        Tests options reset
        """
        options = Options()
        options_save = copy.copy(options)
        options.minimizers = {}
        options.software = ['updated_software1', 'updated_software2']

        options.reset()
        self.assertDictEqual(options.__dict__, options_save.__dict__)


if __name__ == '__main__':
    unittest.main()



"""
Set up performance profiles for both accuracy and runtime tables
"""
from collections import OrderedDict
import os
from textwrap import wrap

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy as np


def profile(results, fig_dir):
    """
    Function that generates profiler plots

    :param results: results nested array of objects
    :type results: list of list of
                    fitbenchmarking.utils.fitbm_result.FittingResult
    :param fig_dir: path to directory containing the figures
    :type fig_dir: str

    :return: path to acc and runtime profile graphs
    :rtype: tuple(str, str)
    """
    acc_bound, runtime_bound = prepare_profile_data(results)
    plot_path = plot(acc_bound, runtime_bound, fig_dir)
    return plot_path


def prepare_profile_data(results):
    """
    Helper function which generates acc and runtime dictionaries which
    contain number of occurrences that the minimizer produces a normalised
    result which is less than the bounds in PROFILER_BOUNDS

    :param results: results nested array of objects
    :type results: list of list of
                   fitbenchmarking.utils.fitbm_result.FittingResult

    :return: dictionary containing number of occurrences
    :rtype: tuple(dict, dict)
    """
    out_acc = []
    out_runtime = []
    for res in results:
        out_acc.append([r.norm_acc for r in res])
        out_runtime.append([r.norm_runtime for r in res])
    minimizers = [r.minimizer for r in results[0]]

    acc_array = np.array(out_acc).T
    runtime_array = np.array(out_runtime).T
    acc_dict = OrderedDict()
    runtime_dict = OrderedDict()

    for i, m in enumerate(minimizers):
        acc_dict[m] = acc_array[i][:]
        runtime_dict[m] = runtime_array[i][:]
    return acc_dict, runtime_dict


def plot(acc, runtime, fig_dir):
    """
    Function that generates profiler plots

    :param acc: acc dictionary containing number of occurrences
    :type acc: dict
    :param runtime: runtime dictionary containing number of occurrences
    :type runtime: dict
    :param fig_dir: path to directory containing the figures
    :type fig_dir: str

    :return: path to acc and runtime profile graphs
    :rtype: tuple(str, str)
    """
    figure_path = []
    for profile_plot, name in zip([acc, runtime], ["acc", "runtime"]):
        this_filename = os.path.join(fig_dir, "{}_profile.png".format(name))
        figure_path.append(this_filename)

        step_values = []
        for value in profile_plot.values():
            sorted_list = np.sort(value)
            step_values.append(np.insert(sorted_list, 0, 0.0))

        max_value = np.max([np.max(v)
                            for v in profile_plot.values()])
        linear_upper_limit = 10

        use_log_plot = True
        if max_value < linear_upper_limit:
            # if we don't need a log plot, then don't print one
            fig, ax = plt.subplots(1, 2,
                                   gridspec_kw={
                                       'width_ratios': [30, 7.8],
                                       'wspace': 0.01})
            legend_ax = 1
            use_log_plot = False
        else:
            fig, ax = plt.subplots(1, 3,
                                   sharey=True,
                                   gridspec_kw={
                                       'width_ratios': [10, 20, 7.8],
                                       'wspace': 0.01})
            legend_ax = 2

        # Plot linear performance profile
        create_plot(ax[0], step_values, acc.keys())
        ax[0].set_xlim(1, linear_upper_limit)
        ax[0].set_xticks([1, 2, 4, 6, 8, 10])
        ax[0].set_xticklabels(['$1$', '$2$', '$4$', '$6$', '$8$', '$10$'])
        ax[0].yaxis.set_ticks_position('left')

        if use_log_plot:
            # Plot log performance profile
            create_plot(ax[1], step_values, acc.keys())
            ax[1].set_xlim(
                linear_upper_limit,
                min(max_value+1, 10000))
            ax[1].set_xscale('log')
            ax[1].set_xticks([100, 1000, 10000])
            ax[1].set_xticklabels(['$10^2$', '$10^3$', '$10^4$'])
            ax[1].yaxis.set_ticks_position('right')
            ax[1].tick_params(axis='y', labelcolor='white')

        # legend
        ax[legend_ax].axis('off')
        handles, labels = ax[0].get_legend_handles_labels()
        wrapped_labels = ['\n'.join(wrap(l, 22)) for l in labels]
        ax[legend_ax].legend(handles, wrapped_labels, loc=2, prop={'size': 7})

        # Common parts
        plt.ylim(0.0, 1.0)
        fig.suptitle("Performance profile - {}".format(name))

        # add a big axis, hide frame
        fig.add_subplot(111, frameon=False)
        # hide tick and tick label of the big axis
        plt.tick_params(labelcolor='none',
                        top=False,
                        bottom=False,
                        left=False,
                        right=False)
        # the xlabel has added white space to (empirically) make it look
        # centred under the plots (and not the legend)
        plt.xlabel("f                       ")
        plt.ylabel("fraction for which solver within f of best")
        ax[0].set_ylim(0.0, 1.05)

        plt.savefig(this_filename, dpi=150)

    return figure_path


def create_plot(ax, step_values, solvers):
    """
    Function to draw the profile on a matplotlib axis

    :param ax: A matplotlib axis to be filled
    :type ax: an `.axes.SubplotBase` subclass of `~.axes.Axes`
              (or a subclass of `~.axes.Axes`)
    :param step_values: a sorted list of the values of the metric
                        being profiled
    :type step_values: list of float
    :param solvers: A list of the labels for the different solvers
    :type solvers: list of strings
    """

    lines = ["-", "-.", "--", ":"]
    # use only 9 of matplotlib's colours, as this will give us
    # 9 * 4 = 36 line/colour combinations, as opposed to
    # 10 * 4 / 2 = 20 if we used all 10
    colors = ["C0", "C1", "C2", "C3", "C4", "C5", "C6", "C7", "C8"]

    huge = 1.0e20  # set a large value as a proxy for infinity
    plot_points = np.linspace(0.0, 1.0, step_values[0].size)
    plot_points = np.append(plot_points, 1.0)

    for i, (solver, solver_values) in enumerate(zip(solvers, step_values)):
        inf_indices = np.where(solver_values > huge)
        solver_values[inf_indices] = huge
        if inf_indices[0].size > 0:
            plural_ending = "s"
            if inf_indices[0].size == 1:
                plural_ending = ""
            solver = "{} ({} failure{})".format(solver,
                                                len(inf_indices[0]),
                                                plural_ending)
        solver_values = np.append(solver_values, huge)
        ax.step(solver_values,
                plot_points,
                label=solver,
                linestyle=lines[(i % len(lines))],
                color=colors[(i % len(colors))],
                lw=2.0,
                where='post')
